package huffman;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Collections;

/**
 * This class contains methods which, when used together, perform the
 * entire Huffman Coding encoding and decoding process
 * 
 * @author Ishaan Ivaturi
 * @author Prince Rawal
 * @author Olivia Duong 192002200
 */
public class HuffmanCoding {
    /**
     * Writes a given string of 1's and 0's to the given file byte by byte
     * and NOT as characters of 1 and 0 which take up 8 bits each
     * 
     * @param filename The file to write to (doesn't need to exist yet)
     * @param bitString The string of 1's and 0's to write to the file in bits
     */
    public static void writeBitString(String filename, String bitString) {
        byte[] bytes = new byte[bitString.length() / 8 + 1];
        int bytesIndex = 0, byteIndex = 0, currentByte = 0;

        // Pad the string with initial zeroes and then a one in order to bring
        // its length to a multiple of 8. When reading, the 1 signifies the
        // end of padding.
        int padding = 8 - (bitString.length() % 8);
        String pad = "";
        for (int i = 0; i < padding-1; i++) pad = pad + "0";
        pad = pad + "1";
        bitString = pad + bitString;

        // For every bit, add it to the right spot in the corresponding byte,
        // and store bytes in the array when finished
        for (char c : bitString.toCharArray()) {
            if (c != '1' && c != '0') {
                System.out.println("Invalid characters in bitstring");
                System.exit(1);
            }

            if (c == '1') currentByte += 1 << (7-byteIndex);
            byteIndex++;
            
            if (byteIndex == 8) {
                bytes[bytesIndex] = (byte) currentByte;
                bytesIndex++;
                currentByte = 0;
                byteIndex = 0;
            }
        }
        
        // Write the array of bytes to the provided file
        try {
            FileOutputStream out = new FileOutputStream(filename);
            out.write(bytes);
            out.close();
        }
        catch(Exception e) {
            System.err.println("Error when writing to file!");
        }
    }
    
    /**
     * Reads a given file byte by byte, and returns a string of 1's and 0's
     * representing the bits in the file
     * 
     * @param filename The encoded file to read from
     * @return String of 1's and 0's representing the bits in the file
     */
    public static String readBitString(String filename) {
        String bitString = "";
        
        try {
            FileInputStream in = new FileInputStream(filename);
            File file = new File(filename);

            byte bytes[] = new byte[(int) file.length()];
            in.read(bytes);
            in.close();
            
            // For each byte read, convert it to a binary string of length 8 and add it
            // to the bit string
            for (byte b : bytes) {
                bitString = bitString + 
                String.format("%8s", Integer.toBinaryString(b & 0xFF)).replace(' ', '0');
            }

            // Detect the first 1 signifying the end of padding, then remove the first few
            // characters, including the 1
            for (int i = 0; i < 8; i++) {
                if (bitString.charAt(i) == '1') return bitString.substring(i+1);
            }
            
            return bitString.substring(8);
        }
        catch(Exception e) {
            System.out.println("Error while reading file!");
            return "";
        }
    }

    /**
     * Reads a given text file character by character, and returns an arraylist
     * of CharFreq objects with frequency > 0, sorted by frequency
     * 
     * @param filename The text file to read from
     * @return Arraylist of CharFreq objects, sorted by frequency
     */
    public static ArrayList<CharFreq> makeSortedList(String filename) {
        StdIn.setFile(filename);
        //////////////////////////////////////////////////////* Your code goes here *///////////////////////////////////////////////////////
        
        int[] array = new int[128]; // array of individual char frequency
        int total = 0; //total of chars

        while (StdIn.hasNextChar()) {

            char f = StdIn.readChar();
            array[f]+=1;
            total=total+1; //incrementing
        }
        
        ArrayList<CharFreq> newlist = new ArrayList<>();  
        
        for (int i = 0; i < 128; i=i+1) {
            if (array[i] != 0){
                double prob = (double) array[i]/total;
                CharFreq f = new CharFreq((char)i, prob);
                newlist.add(f);
            }
        }
        
        Collections.sort(newlist);

        if (newlist.size() == 1){
            if (array[127] != 0){
                newlist.add(new CharFreq((char)0, 0));
            }
            else{
                char f2 = newlist.get(newlist.size() - 1).getCharacter();
                newlist.add(new CharFreq((char)((int)f2 + 1), 0));
            }
        }
        Collections.sort(newlist);
        return newlist;
    }

        //return null; // Delete this line
    
    
    ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    /**
     * Uses a given sorted arraylist of CharFreq objects to build a huffman coding tree
     * 
     * @param sortedList The arraylist of CharFreq objects to build the tree from
     * @return A TreeNode representing the root of the huffman coding tree
     */
    public static TreeNode makeTree(ArrayList<CharFreq> sortedList) {
        /* Your code goes here */
        
        Queue<TreeNode> aim = new Queue<>();
        Queue<TreeNode> original = new Queue<>();
       

        for (int  i = 0; i < sortedList.size(); i+=1){

            original.enqueue(new TreeNode(sortedList.get(i), null, null));
            
        }

        TreeNode right = null;
        TreeNode left = null;
       
        double p = 0;

        while (!original.isEmpty()){

            if (aim.isEmpty()){

                left = original.dequeue();
                right = original.dequeue();
            }
            
            else{
                
                if (original.peek().getData().getProbOccurrence() <= aim.peek().getData().getProbOccurrence()){

                    left = original.dequeue();

                }
                else left = aim.dequeue();

                
                if (!original.isEmpty() && !aim.isEmpty()){

                    if (original.peek().getData().getProbOccurrence() <= aim.peek().getData().getProbOccurrence()){
                        
                        right = original.dequeue();
                    }

                    else right = aim.dequeue();
                }

                
                else if (original.isEmpty() && !aim.isEmpty()){

                    right = aim.dequeue();

                } 
                else right = original.dequeue(); //if empty
            }
            
            p = left.getData().getProbOccurrence() + right.getData().getProbOccurrence();
            aim.enqueue(new TreeNode(new CharFreq(null, p), left, right));
        }
        
        while (aim.size() > 1){

            left = aim.dequeue();
            right = aim.dequeue();
            p = left.getData().getProbOccurrence() + right.getData().getProbOccurrence();
            aim.enqueue(new TreeNode(new CharFreq(null, p), left, right));

        }

        return aim.dequeue();
        //return null; // Delete this line
    }

     ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Uses a given huffman coding tree to create a string array of size 128, where each
     * index in the array contains that ASCII character's bitstring encoding. Characters not
     * present in the huffman coding tree should have their spots in the array left null
     * 
     * @param root The root of the given huffman coding tree
     * @return Array of strings containing only 1's and 0's representing character encodings
     */
  
    public static String[] makeEncodings(TreeNode root) {
        /* Your code goes here */

        String[] encodings = new String[128];
        encodingCode(encodings, root, "");
        return encodings;
    }

    private static void encodingCode(String[] encoded, TreeNode root, String code) {

        if (root.getLeft() != null) { //left add 0
            
            encodingCode(encoded, root.getLeft(), code + "0");

        }

        if (root.getRight() != null){ //right add 1

            encodingCode(encoded, root.getRight(), code + "1");

        }
        
        if (root.getLeft() == null && root.getRight() == null){

            encoded[root.getData().getCharacter()] = code;
            return;

        }

        return;
    }
        //return null; // Delete this line

    
    ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Using a given string array of encodings, a given text file, and a file name to encode into,
     * this method makes use of the writeBitString method to write the final encoding of 1's and
     * 0's to the encoded file.
     * 
     * @param encodings The array containing binary string encodings for each ASCII character
     * @param textFile The text file which is to be encoded
     * @param encodedFile The file name into which the text file is to be encoded
     */
    public static void encodeFromArray(String[] encodings, String textFile, String encodedFile) {

        StdIn.setFile(textFile);
        /* Your code goes here */

        String data = new String();   //initialize 

        while (StdIn.hasNextChar()){

            char character = StdIn.readChar();
            data = data + encodings[character];
            
        }
      
        for (int i = 0; i < data.length(); i=i+1){

            if (data.charAt(i) != '1' && data.charAt(i) != '0'){

                System.out.println(data.charAt(i));

            }
        }

        writeBitString(encodedFile, data);
        
        return;

    }
    //ugh why is this sooo hard sadlkfjsdklfjeiljipjcxkjsdr

    ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Using a given encoded file name and a huffman coding tree, this method makes use of the 
     * readBitString method to convert the file into a bit string, then decodes the bit string
     * using the tree, and writes it to a file.
     * 
     * @param encodedFile The file which contains the encoded text we want to decode
     * @param root The root of your Huffman Coding tree
     * @param decodedFile The file which you want to decode into
     */
    public static void decode(String encodedFile, TreeNode root, String decodedFile) {

        StdOut.setFile(decodedFile);
        /* Your code goes here */

        StdOut.setFile(decodedFile);
        String encoding = readBitString(encodedFile);
        TreeNode pointer = root;
        int o = 0; 

        while (o < encoding.length()){

            if ((encoding.charAt(o) == '1') && (pointer.getRight()) != null){

                pointer = pointer.getRight();
    
            }


            if ((encoding.charAt(o) == '0') && (pointer.getLeft()) != null){

                pointer = pointer.getLeft();

            }


            if ((pointer.getLeft() == null) || (pointer.getRight() == null)){

                StdOut.print(pointer.getData().getCharacter()); 
                pointer = root;

            }
            o=o+1;
        }
    }
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////